package qsnake;

import javax.swing.*;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Toolkit;
import java.net.URL;
import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;
import qsnake.Snake.Direction;

public class Board extends JPanel{
    public static int speed = 75;
    private static int size = 0;
    
    private int episode = 0;
    
    
    private final int BOARD_WIDTH = 640;
    private final int BOARD_HEIGHT = 640;
    private final int SINGLE_SIZE = 32;
    
    private final int N_FRUITS = 3;
    
    private Image tail;
    private Image head;
    private Image fruits[] = new Image[N_FRUITS];
    
    private Snake snake;
    
    private int fruit_x = 32;
    private int fruit_y = 0;
    
    private Direction previous = Direction.DOWN;
    
    private QLearning selector = new QLearning();
    
    public Board() {
        initBoard();
        snake = new Snake(BOARD_WIDTH, BOARD_HEIGHT);
        System.err.println("Episode: " + episode);
    }
    
    private void initBoard() {
        setBackground(Color.black);
        setFocusable(true);

        setPreferredSize(new Dimension(BOARD_WIDTH, BOARD_HEIGHT));
        loadImages();
        
        selector.printTable();
    } 
    
    
    private void loadImages() {
        URL url = this.getClass().getResource("/images/head.png");
        ImageIcon icon_h = new ImageIcon(url);
        head = icon_h.getImage();
        
        url = this.getClass().getResource("/images/tail.png");
        ImageIcon icon_t = new ImageIcon(url);
        tail = icon_t.getImage();

        for(int i = 0; i < N_FRUITS; i++){        
            url = this.getClass().getResource("/images/apple.png");
            ImageIcon icon_f = new ImageIcon(url);
            fruits[i] = icon_f.getImage();
        }

    }
    
    @Override
    public void paintComponent(Graphics g) {
        super.paintComponent(g);

        doDrawing(g);
    }
    
    
    private void doDrawing(Graphics g) {
        try {
            g.drawImage(head, snake.xhead(), snake.yhead(), this);
            
            for(int i = 1; i < snake.size(); i++) {
                g.drawImage(tail, snake.xtail(i), snake.ytail(i), this);
            }
            
            g.drawImage(fruits[0], fruit_x, fruit_y, this);
            Toolkit.getDefaultToolkit().sync();
            
            Direction dir;
            
            
            State oldState = getSnakeState();
            dir = selector.select(oldState);
            
            if(snake.xhead() == fruit_x && snake.yhead() == fruit_y) {
                do{
                    fruit_x = new Random().nextInt(20) * SINGLE_SIZE;
                    fruit_y = new Random().nextInt(20) * SINGLE_SIZE;
                }while(snake.fruitCollision(fruit_x, fruit_y));
                repaint();
                snake.eat(dir);
            }
            else
            {
                snake.move(dir);
                
                int currentSize = snake.size();
                
                DetailsWindow.currentSizeLabel.setText("Current Size: " + currentSize);
                if(currentSize > size) {
                    DetailsWindow.currentSizeLabel.setForeground(Color.red);
                }
                
            }
            
            Thread.sleep(speed);
            
            switch(dir)
            {
                case UP: previous = Direction.DOWN; break;
                case RIGHT: previous = Direction.LEFT; break;
                case DOWN: previous = Direction.UP; break;
                case LEFT: previous = Direction.RIGHT; break;
            }
            
            repaint();
            if(snake.isDead()) {
                if(size < snake.size())
                {
                    size = snake.size();
                    System.err.println("NUOVO RECORD! " + size);
                    
                    DetailsWindow.recordSizeLabel.setText("Record Size: " + size);
                }
                
                selector.update(oldState, dir);
                snake = new Snake(BOARD_WIDTH, BOARD_HEIGHT);
                
                System.err.println("Episode: " + ++episode + " (" + size + ")");
                DetailsWindow.episodeLabel.setText("Episode: " + episode);
                
                previous = Direction.DOWN;
                DetailsWindow.clear(previous);
                
                DetailsWindow.currentSizeLabel.setText("Current Size: 1");
                DetailsWindow.currentSizeLabel.setForeground(Color.black);
            }
            else{
                State snakeState = getSnakeState();
                selector.update(oldState, snakeState, dir);
                
                DetailsWindow.clear(previous);
                DetailsWindow.draw(snakeState, previous);
            }
            
        } catch (Exception ex) {
            Logger.getLogger(Board.class.getName()).log(Level.SEVERE, null, ex);
        }
    }
    
    private State getSnakeState() {
        int[] state = new int[8];
                
        for(int i = 1; i < snake.size(); i++) {
            if(snake.xhead() + SINGLE_SIZE == snake.xtail(i) && snake.yhead() == snake.ytail(i)
                    || snake.xhead() + SINGLE_SIZE >= BOARD_WIDTH)
            {
                state[1] = 1;
            }
            
            if(snake.xhead() - SINGLE_SIZE == snake.xtail(i) && snake.yhead() == snake.ytail(i) 
                    || snake.xhead() - SINGLE_SIZE < 0)
            {
                state[3] = 1;
            }
            
            if(snake.yhead() + SINGLE_SIZE == snake.ytail(i) && snake.xhead() == snake.xtail(i) 
                    || snake.yhead() + SINGLE_SIZE >= BOARD_HEIGHT)
            {
                state[2] = 1;
            }
            
            if(snake.yhead() - SINGLE_SIZE == snake.ytail(i) && snake.xhead() == snake.xtail(i) 
                    || snake.yhead() - SINGLE_SIZE < 0)
            {
                state[0] = 1;
            }
        }
        
        state[previous.ordinal()] = -1;
        
        if(fruit_x > snake.xhead()) {
            state[5] = 1;
        }
        else if(fruit_x < snake.xhead()) {
            state[7] = 1;
        }
        
        if(fruit_y > snake.yhead()) {
            state[6] = 1;
        }
        else if(fruit_y < snake.yhead()) {
            state[4] = 1;
        }
        
        return new State(state);
    }
    
    
}
