package qsnake;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.Random;
import javafx.util.Pair;
import qsnake.Snake.Direction;

public class QLearning {    
    private float randomize = 70;
             
    HashMap<Pair<String, Direction>, Float> table = new HashMap();
    HashMap<Pair<String, Direction>, Float> elegibility = new HashMap();
    
    private ArrayList<State> states = new ArrayList();
    
    private float gamma = 0.25f, alpha = 0.8f, lambda = 0.8f;
       
    public QLearning() {
    }
    
    public void printState(State state) {
        for(int i = 0; i < state.size() - 1; i++) {
            System.out.print(state.get(i) + ",");
        }
        
        System.out.print(state.get(state.size()-1));
    }
    
    public void printTable() {
        System.out.println(table.size() + states.size());
        
        table.forEach((pair, value) -> {
                System.out.print(pair.getKey() + " ");
                System.out.print(pair.getKey()); 
                System.out.println(" " + pair.getValue() + "\t" + value);
        });
    }
    
    public Direction select(State state) {        
        addState(state);
        
        if(new Random().nextInt(1000) < randomize){
            randomize -= randomize*0.2;
            System.err.println("RANDOM: " + randomize);
            
            if(randomize < 2)
                randomize = 0;
            
            return greedyAction(state);
        }
        
        return maxAction(state);
    }
    
    void update(State oldState, State newState, Direction action) {
        Pair <String, Direction> p = new Pair(oldState.toString(), action);
        
        
        for(Map.Entry<Pair<String, Direction>, Float> e : elegibility.entrySet()) {
            e.setValue(e.getValue() * gamma * lambda);
        }
        
        elegibility.replace(p, 1f);
        
        Float value = getReward(oldState, action) + gamma * table.get(new Pair(newState.toString(), maxAction(newState)));
        
        for(Map.Entry<Pair<String, Direction>, Float> t : table.entrySet()) {
            t.setValue(
                t.getValue() + alpha * elegibility.get(t.getKey()) * (value - t.getValue())    
            );  
        }
    }
    
    void update(State oldState, Direction action) {
        Pair <String, Direction> p = new Pair(oldState.toString(), action);
                
        table.replace(p, table.get(p) + getReward(oldState, action));
        
        for(Map.Entry<Pair<String, Direction>, Float> e : elegibility.entrySet()) {
            e.setValue(0f);
        }
    }
        
    public Direction maxAction(State state) {
        Float reward = null;
        Direction action = null;
        
        addState(state);
        
        for(Pair<String, Direction> p : table.keySet()) {
            if(p.getKey().compareTo(state.toString()) == 0) {
                if(reward == null || reward < table.get(p)) {
                    action = p.getValue();
                    reward = table.get(p);
                }
            }  
        }
        
        return action;
    }
    
    public Direction randomAction(State state) {
        addState(state);
        
        int r = new Random().nextInt(3);
        
        for(Pair<String, Direction> p : table.keySet()) {
            if(p.getKey().compareTo(state.toString()) == 0 && r == 0) {
                return p.getValue();
            }
            else if(p.getKey().compareTo(state.toString()) == 0 && r == 1) {
                return p.getValue();
            }
            else if(p.getKey().compareTo(state.toString()) == 0 && r == 2) {
                return p.getValue();
            } 
        }
        
        return null;
    }
    
    public Direction greedyAction(State state) {
        addState(state);
        
        Direction action = null;
        Float max = 0f;
        
        int r = new Random().nextInt(3);
        
        for(Pair<String, Direction> p : table.keySet()) {
            if(p.getKey().compareTo(state.toString()) == 0) {
                if(action == null || getReward(state, p.getValue()) > max)
                {
                    max = getReward(state, p.getValue());
                    action = p.getValue();
                }
            }
        }
        
        return action;
    }
    
    private void addState(State state) {        
        for(Pair<String, Direction> p : table.keySet()) {
            if(p.getKey().compareTo(state.toString()) == 0) {
                return; 
            }
        }
                
        for(int k = 0; k < 4; k++) {
            if(state.get(k) != -1) {
                Direction dir = Direction.valueOf(k);
                Pair<String, Direction> pair = new Pair(state.toString(), dir);
                if(!table.containsKey(pair)){
                    table.put(pair, 0f);
                    elegibility.put(pair, 0f); 
                    
                    DetailsWindow.drawLast(state);
                }
            }
        }
    }
    
    private Float getReward(State state, Direction dir) {
        if(state.get(dir.ordinal()) == 1)
            return -5f;
        
        if(state.get(dir.ordinal() + 4) == 1)
            return 2f; 
        
        return -0.5f;
    }
}
