package qsnake;

public class Snake {
    public enum Direction {
        UP(0), RIGHT(1), DOWN(2), LEFT(3);
        
        private int value;
    
        private Direction(int value) {
            this.value = value;
        }
        
        public static Direction valueOf(int dir) {
            switch(dir)
            {
                case 0: return UP;
                case 1: return RIGHT;
                case 2: return DOWN;
                case 3: return LEFT;
            }
            
            return UP;
        }
    };
    
    private final int MAX_LENGTH;
    private final int SINGLE_SIZE = 32;
    
    private final int BOARD_WIDTH;
    private final int BOARD_HEIGHT;
    
    private int body_x[];
    private int body_y[];
    private int size = 3;
    
    public Snake(int width, int height) {
        BOARD_WIDTH = width;
        BOARD_HEIGHT = height;
        
        MAX_LENGTH = (width*height)/SINGLE_SIZE;
        body_x = new int[MAX_LENGTH];
        body_y = new int[MAX_LENGTH];
        
        for(int i = 0; i < size; i++) {
            body_x[i] = SINGLE_SIZE;
            body_y[i] = height - (SINGLE_SIZE * (size+1)) + SINGLE_SIZE*i;
        }
    }
    
    public void eat(Direction d) {
        body_x[size] = body_x[size-1];
        body_y[size] = body_y[size-1];
        
        move(d);
        
        size++;
        
    }
        
    public void move(Direction d) {
        for(int i = size-1; i > 0; i--) {
            body_x[i] = body_x[i-1];
            body_y[i] = body_y[i-1];
        }
        
        switch(d) {
            case UP:
                body_y[0] -= SINGLE_SIZE;
                break;
            case RIGHT:
                body_x[0] += SINGLE_SIZE;
                break;
            case LEFT:
                body_x[0] -= SINGLE_SIZE;
                break;
            case DOWN:
                body_y[0] += SINGLE_SIZE;
                break;
        }
    }
    
    private boolean boardCollision() {
        return xhead() < 0 || xhead() >= BOARD_WIDTH || 
                yhead() < 0 || yhead() >= BOARD_HEIGHT; 
    }
    
    private boolean bodyCollision() {
        for(int i = 1; i < size; i++) {
            if(xhead() == xtail(i) && yhead() == ytail(i))
                return true;
        }
        
        return false;
    }
    
    public boolean fruitCollision(int x, int y) {
        for(int i = 0; i < size; i++) {
            if(body_x[i] == x && body_y[i] == y)
                return true;
        }
        
        return false;
    }
    
    public boolean isDead() {
        return boardCollision() || bodyCollision();
    }
    
    public int xhead() {
        return body_x[0];
    }
    
    public int yhead() {
        return body_y[0];
    }
    
    public int xtail(int i) {
        return body_x[i];
    }
    
    public int ytail(int i) {
        return body_y[i];
    }
    public int size() {
        return size;
    }
    
    
    
}
